from typing import Union, Iterable, Tuple

import numpy as np

from elizur.life.annuity import discount_factor
from elizur.life.util import validate_age, validate_interval, validate_t_interval


class LifeTable:
    # pylint: disable=too-many-public-methods
    # pylint: disable=too-many-instance-attributes
    """
    Given an input tuple of failure probabilities this class is capable
    of calculating common actuarial functions such as qx, px, lx, dx,
    ex, mx, w, nqx, npx, nlx, ndx, nmx, tqxn, Dx, Sx, Nx, Cx, Mx, Rx, actuarial
    present value of annuities, and actuarial present value of insurances.

    Args:
        table: iterable of failure probabilities as floats in
               sequential order, e.g. (1q0, 2q1, ..., 100q99)
        initial_pop: the size of the initial population (l0)
    """

    def __init__(
        self,
        table: Union[Iterable, np.array],
        name: str = "",
        description: str = "",
        initial_pop: int = 100000,
    ):
        self.qxs = np.array(table)
        self.table_size = self.qxs.size
        self.pxs = 1 - self.qxs
        self.lxs = self._set_lxs(initial_pop)
        self.dxs = -1 * np.diff(self.lxs)
        self.mxs = np.divide(self.dxs, self.lxs[:-1])
        self.name = name
        self.description = description

    def _set_lxs(self, l0: int) -> Tuple[float]:
        """
        Args:
            l0: the size of the initial population
        """
        return l0 * np.insert(np.cumprod(self.pxs), 0, 1)

    @property
    def w(self) -> int:
        """
        Returns:
            The limiting age of the life table
        """
        return self.get_lxs().index(0) - 1

    @validate_age
    def qx(self, x: int) -> float:
        """
        Args:
            x: start age

        Returns:
            The probability of failure between the ages x and x + 1
        """
        if x >= self.table_size:
            return 1.0
        return self.qxs[x]

    @validate_age
    def px(self, x: int) -> float:
        """
        Args:
            x: start age

        Returns:
            The probability of survival between the ages x and x + 1
        """
        if x >= self.table_size:
            return 0.0
        return self.pxs[x]

    @validate_age
    def lx(self, x: int) -> float:
        """
        Args:
            x: age

        Returns:
            The population size at time x.  This is the same thing as the
            number of person years lived between age x and x + 1.
        """
        if x >= self.table_size:
            return 0.0
        return self.lxs[x]

    @validate_age
    def dx(self, x: int) -> float:
        """
        Args:
            x: start age

        Returns:
            The number of failures between ages x and x + 1
        """
        if x >= self.table_size:
            return 0.0
        return self.dxs[x]

    @validate_age
    def mx(self, x: int) -> float:
        """
        Args:
            x: start age
        Returns:
            The central failure rate between ages x and x + 1
        """
        if x >= self.table_size:
            return 1.0
        return self.mxs[x]

    @validate_age
    def ex(self, x: int) -> float:
        """
        Args:
            x: start age

        Returns:
            The curtate life expectation at age x
        """
        if x >= self.table_size:
            return 0.0
        return sum([self.lx(t + x) / self.lx(x) for t in range(1, self.w - x)])

    @validate_age
    @validate_interval
    def nqx(self, n: int, x: int) -> float:
        """
        Args:
            n: width of failure interval in years
            x: start age

        Returns:
            The probability of failure between ages x and x + n
        """
        if x >= self.table_size:
            return 1.0
        return (self.lx(x) - self.lx(n + x)) / self.lx(x)

    @validate_interval
    def nqxs(self, n: int) -> np.array:
        """
        Args:
            n: width of failure interval in years

        Returns:
            The probability of failure between ages x and x + n for all ages
        """
        return np.array(
            [
                (self.lx(x) - self.lx(x + n)) / self.lx(x)
                for x in range(self.table_size)
                if x + n < self.table_size
            ]
        )

    @validate_age
    @validate_interval
    def npx(self, n: int, x: int) -> float:
        """
        Args:
            n: width of survival interval in years
            x: start age

        Returns:
            The probability of survival between ages x and x + n
        """
        if x >= self.table_size:
            return 0.0
        return self.lx(n + x) / self.lx(x)

    @validate_interval
    def npxs(self, n: int) -> np.array:
        """
        Args:
            n: width of survival interval in years

        Returns:
            The probability of survival between ages x and x + n for all ages
        """
        return np.array(
            [
                self.lx(x + n) / self.lx(x)
                for x in range(self.table_size)
                if x + n < self.table_size
            ]
        )

    @validate_age
    @validate_interval
    def nlx(self, n: int, x: int) -> float:
        """
        Args:
            n: width of failure interval in years
            x: start age

        Returns:
            The number of person years lived between ages x and x + n
        """
        return sum([self.lx(x + i) for i in range(n)])

    @validate_age
    @validate_interval
    def ndx(self, n: int, x: int) -> float:
        """
        Args:
            n: width of failure interval in years
            x: start age

        Returns:
            The number of failures between ages x and x + n
        """
        return sum([self.dx(x + i) for i in range(n)])

    @validate_age
    @validate_interval
    def nmx(self, n: int, x: int) -> float:
        """
        Args:
            x: start age

        Returns:
            The central failure rate between ages x and x + n
        """
        if x >= self.table_size:
            return 1.0
        return self.ndx(n, x) / self.nlx(n, x)

    @validate_age
    @validate_interval
    @validate_t_interval
    def tqxn(self, t: int, n: int, x: int) -> float:
        """
        Args:
            t: width of the failure interval in years
            n: width of the survival interval in years
            x: start age

        Returns:
            The probability of surviving from age x to x + n and
            then failing between age x + n and age x + n + t
        """
        if x >= self.table_size:
            return 1.0
        # probability of surviving past max age is 0
        if x + n >= self.table_size:
            return 0.0
        # probability of failing after max age given
        # survival to less than equal to max age is 1
        if x + n + t >= self.table_size:
            return 1.0
        return self.npx(n, x) * self.nqx(t, x + n)

    def tqxns(self, t: int, n: int) -> np.array:
        """
        Args:
            t: width of the failure interval in years
            n: width of the survival interval in years

        Returns:
            The probability of surviving from age x to x + n and
            then failing between age x + n and age x + n + t for all
            ages
        """
        if n <= 0 or t <= 0:
            return np.zeros(self.table_size)
        return np.array(
            [
                self.npx(n, x) * self.nqx(t, x + n)
                for x in range(self.table_size)
                if x + n + t < self.table_size
            ]
        )

    def get_qxs(self) -> Tuple[float]:
        """
        This method is deprecated and will be removed in v1.0.0

        The recommended method is to use the 'qxs' property

        Returns:
            A tuple of the failure
            probabilities (qxs), e.g., (0q1, 2q1, ..., 100q99)
        """
        return tuple(self.qxs)

    def get_pxs(self) -> Tuple[float]:
        """
        This method is deprecated and will be removed in v1.0.0

        The recommended method is to use the 'pxs' property

        Returns:
            A tuple of the survival
            probabilities (pxs), e.g., (0p1, 2p1, ..., 100p99)
        """
        return tuple(self.pxs)

    def get_lxs(self) -> Tuple[float]:
        """
        This method is deprecated and will be removed in v1.0.0

        The recommended method is to use the 'lxs' property

        Returns:
            A tuple of the population counts (lxs)
        """
        return tuple(self.lxs)

    @validate_age
    def Dx(self, x: int, i: float) -> float:
        """
        Actuarial commutation function Dx

        Args:
            x: start age
            i: interest rate
        Returns:
            Population at age x discounted for x years
        """
        return self.lx(x) * discount_factor(i) ** x

    @validate_age
    def Nx(self, x: int, i: float) -> float:
        """
        Actuarial commutation function Nx

        Args:
            x: start age
            i: interest rate
        Returns:
            Sum of Ds from age x and onward
        """
        return sum([self.Dx(n, i) for n in range(x, self.w + 1)])

    @validate_age
    def Sx(self, x: int, i: float) -> float:
        """
        Actuarial commutation function Sx

        Args:
            x: start age
            i: interest rate
        Returns:
            Sum of Ns from age x and onward
        """
        return sum([self.Nx(n, i) for n in range(x, self.w + 1)])

    @validate_age
    def Cx(self, x: int, i: float) -> float:
        """
        Actuarial commutation function Cx

        Args:
            x: start age
            i: interest rate
        Returns:
            Failures between x and x + 1 discounted for x years
        """
        return (self.lx(x) - self.lx(x + 1)) * discount_factor(i) ** (x + 1)

    @validate_age
    def Mx(self, x: int, i: float) -> float:
        """
        Actuarial commutation function Mx

        Args:
            x: start age
            i: interest rate
        Returns:
            Sum of Cs from age x and onward
        """
        return sum([self.Cx(t, i) for t in range(x, self.w + 1)])

    @validate_age
    def Rx(self, x: int, i: float) -> float:
        """
        Actuarial commutation function Rx

        Args:
            x: start age
            i: interest rate
        Returns:
            Sum of Ms from age x and onward
        """
        return sum([self.Mx(t, i) for t in range(x, self.w + 1)])

    @validate_age
    def Ax(self, x: int, i: float) -> float:
        """
        Args:
            x: start age
            i: interest rate
        Returns:
            Actuarial present value of level whole insurance
        """
        return self.Mx(x, i) / self.Dx(x, i)

    @validate_age
    @validate_interval
    def Axn(self, x: int, i: float, n: int) -> float:
        """
        Args:
            x: start age
            i: interest rate
            n: number of periods in the temporary insurance
        Returns:
            Actuarial present value of level temporary insurance
        """
        return (self.Mx(x, i) - self.Mx(x + n, i)) / self.Dx(x, i)

    @validate_age
    def IAx(self, x: int, i: float) -> float:
        """
        Args:
            x: start age
            i: interest rate
        Returns:
            Actuarial present value of increasing whole insurance
        """
        return self.Rx(x, i) / self.Dx(x, i)

    @validate_age
    @validate_interval
    def IAxn(self, x: int, i: float, n: int) -> float:
        """
        Args:
            x: start age
            i: interest rate
            n: number of periods in the temporary insurance
        Returns:
            Actuarial present value of increasing temporary insurance
        """
        return (self.Rx(x, i) - self.Rx(x + n, i) - n * self.Mx(x + n, i)) / self.Dx(
            x, i
        )

    @validate_age
    def ax(self, x: int, i: float) -> float:
        """
        Args:
            x: start age
            i: interest rate
        Returns:
            Actuarial present value of a level perpetuity
        """
        return self.Nx(x + 1, i) / self.Dx(x, i)

    @validate_age
    @validate_interval
    def axn(self, x: int, i: float, n: int) -> float:
        """
        Args:
            x: start age
            i: interest rate
            n: length of payments
        Returns:
            Actuarial present value of a temporary annuity
        """
        return (self.Nx(x + 1, i) - self.Nx(x + n + 1, i)) / self.Dx(x, i)

    @validate_age
    def ax_due(self, x: int, i: float) -> float:
        """
        Args:
            x: start age
            i: interest rate
        Returns:
            Actuarial present value of a level perpetuity due
        """
        return self.Nx(x, i) / self.Dx(x, i)

    @validate_age
    @validate_interval
    def axn_due(self, x: int, i: float, n: int) -> float:
        """
        Args:
            x: start age
            i: interest rate
            n: length of payments
        Returns:
            Actuarial present value of a temporary annuity due
        """
        return (self.Nx(x, i) - self.Nx(x + n, i)) / self.Dx(x, i)


EXAMPLE_TABLE = (
    0.006271,
    0.00041799999999999997,
    0.00028100000000000005,
    0.000201,
    0.00017,
    0.000163,
    0.000127,
    0.000123,
    0.000133,
    0.000132,
    0.000126,
    0.00013000000000000002,
    0.000145,
    0.000186,
    0.00020999999999999998,
    0.000253,
    0.000389,
    0.00044,
    0.00046600000000000005,
    0.000457,
    0.00045400000000000003,
    0.000502,
    0.000467,
    0.000453,
    0.000486,
    0.000498,
    0.00051,
    0.000507,
    0.000565,
    0.0005989999999999999,
    0.000632,
    0.0006680000000000001,
    0.0007239999999999999,
    0.000786,
    0.000853,
    0.000958,
    0.001034,
    0.0011200000000000001,
    0.001221,
    0.001433,
    0.0014930000000000002,
    0.001653,
    0.00175,
    0.0019950000000000002,
    0.002091,
    0.0023039999999999996,
    0.002376,
    0.002577,
    0.002859,
    0.0030310000000000003,
    0.003194,
    0.003522,
    0.003634,
    0.004142000000000001,
    0.0044340000000000004,
    0.0050999999999999995,
    0.005006,
    0.005886,
    0.006441,
    0.007266,
    0.007575999999999999,
    0.008476000000000001,
    0.009201,
    0.010101,
    0.011149,
    0.012107,
    0.013059,
    0.014570999999999999,
    0.015590999999999999,
    0.017396000000000002,
    0.018991,
    0.020454,
    0.022525,
    0.024633,
    0.027135000000000003,
    0.030098,
    0.032631,
    0.036094,
    0.039472,
    0.044109999999999996,
    0.0493,
    0.053298000000000005,
    0.062179000000000005,
    0.06455,
    0.07505500000000001,
    0.083221,
    0.091996,
    0.10139,
    0.111404,
    0.122037,
    0.13328,
    0.145119,
    0.157532,
    0.170488,
    0.183953,
    0.19788,
    0.21221700000000002,
    0.226905,
    0.241875,
    0.257053,
    1.0,
)



import numbers
from typing import Iterable, Union

import numpy as np


def discount_factor(
    i: Union[float, Iterable, np.array]
) -> Union[float, Iterable, np.array]:
    """
    Args:
        i: interest rate in decimal form
    Returns:
        The related discount factor
    """
    return np.divide(1, np.add(1, i))


def interest_rate(
    d: Union[float, Iterable, np.array]
) -> Union[float, Iterable, np.array]:
    """
    Args:
        d: discount rate in decimal form
    Returns:
        The related interest rate in decimal form
    """
    return np.divide(d, np.subtract(1, d))


def discount_rate(
    i: Union[float, Iterable, np.array]
) -> Union[float, Iterable, np.array]:
    """
    Args:
        i: interest rate in decimal form
    Returns:
        The related discount rate
    """
    return np.divide(i, np.add(1, i))


def annuity_pv(
    n: Union[int, Iterable, np.array], i: Union[float, Iterable, np.array]
) -> Union[float, Iterable, np.array]:
    """
    Args:
        n: years
        i: periodic interest rate in decimal form
    Returns:
        Present value of annuity of n years with an interest rate of i
    """
    return np.divide(1 - np.power(discount_factor(i), n), i)


def annuity_due_pv(
    n: Union[int, Iterable, np.array], i: Union[float, Iterable, np.array]
) -> Union[float, Iterable, np.array]:
    """
    Args:
        n: years
        i: periodic interest rate in decimal form
    Returns:
        Present value of annuity of n years with an interest rate of i
    """
    return np.divide(1 - np.power(discount_factor(i), n), discount_rate(i))


def perpetuity_pv(
    i: Union[float, Iterable, np.array]
) -> Union[float, Iterable, np.array]:
    """
    Args:
        i: periodic interest rate in decimal form
    Returns:
        Present value of a perpetuity with an interest rate of i
    """
    return np.divide(1, i)


def perpetuity_due_pv(
    i: Union[float, Iterable, np.array]
) -> Union[float, Iterable, np.array]:
    """
    Args:
        i: periodic interest rate in decimal form
    Returns:
        Present value of a perpetuity due with an interest rate of i
    """
    return np.divide(1, discount_rate(i))


def annuity_fv(
    n: Union[int, Iterable, np.array], i: Union[float, Iterable, np.array]
) -> Union[float, Iterable, np.array]:
    """
    Args:
        n: years
        i: periodic interest rate in decimal form
    Returns:
        Future value of annuity of n years with an interest rate of i
    """
    return np.multiply(annuity_pv(n, i), np.power(np.add(1, i), n))


def annuity_due_fv(
    n: Union[int, Iterable, np.array], i: Union[float, Iterable, np.array]
) -> Union[float, Iterable, np.array]:
    """
    Args:
        n: years
        i: periodic interest rate in decimal form
    Returns:
        Future value of annuity of n years with an interest rate of i
    """
    return np.multiply(annuity_due_pv(n, i), np.power(np.add(1, i), n))


def increasing_annuity_pv(
    n: Union[int, Iterable, np.array], i: Union[float, Iterable, np.array]
) -> Union[float, Iterable, np.array]:
    """
    Args:
        n: years
        i: periodic interest rate in decimal form
    Returns:
        Present value of an increasing annuity of n years with
        an interest rate of i.  It is assumed that the annuity payment
        increments by 1 each period.
    """
    return np.divide(
        annuity_due_pv(n, i) - np.multiply(n, np.power(discount_factor(i), n)), i
    )


def increasing_annuity_fv(
    n: Union[int, Iterable, np.array], i: Union[float, Iterable, np.array]
) -> Union[float, Iterable, np.array]:
    """
    Args:
        n: years
        i: periodic interest rate in decimal form
    Returns:
        Future value of an increasing annuity of n years with
        an interest rate of i.  It is assumed that the annuity payment
        increments by 1 each period.
    """
    return np.divide(annuity_due_fv(n, i) - n, i)


def increasing_annuity_due_pv(
    n: Union[int, Iterable, np.array], i: Union[float, Iterable, np.array]
) -> Union[float, Iterable, np.array]:
    """
    Args:
        n: years
        i: periodic interest rate in decimal form
    Returns:
        Present value of an increasing annuity due of n years with
        an interest rate of i.  It is assumed that the annuity payment
        increments by 1 each period.
    """
    return np.multiply(np.add(1, i), increasing_annuity_pv(n, i))


def increasing_annuity_due_fv(
    n: Union[int, Iterable, np.array], i: Union[float, Iterable, np.array]
) -> Union[float, Iterable, np.array]:
    """
    Args:
        n: years
        i: periodic interest rate in decimal form
    Returns:
        Future value of an increasing annuity due of n years with
        an interest rate of i.  It is assumed that the annuity payment
        increments by 1 each period.
    """
    return np.multiply(increasing_annuity_due_pv(n, i), np.power(np.add(1, i), n))


def decreasing_annuity_pv(
    n: Union[int, Iterable, np.array], i: Union[float, Iterable, np.array]
) -> Union[float, Iterable, np.array]:
    """
    Args:
        n: years
        i: periodic interest rate in decimal form
    Returns:
        Present value of a decreasing annuity of n years with
        an interest rate of i.  It is assumed that the annuity payment
        decrements by 1 each period.
    """
    return np.divide(n - annuity_pv(n, i), i)


def decreasing_annuity_fv(
    n: Union[int, Iterable, np.array], i: Union[float, Iterable, np.array]
) -> Union[float, Iterable, np.array]:
    """
    Args:
        n: years
        i: periodic interest rate in decimal form
    Returns:
        Future value of a decreasing annuity of n years with
        an interest rate of i.  It is assumed that the annuity payment
        decrements by 1 each period.
    """
    return np.divide(np.multiply(n, np.power(np.add(1, i), n)) - annuity_fv(n, i), i)


def decreasing_annuity_due_pv(
    n: Union[int, Iterable, np.array], i: Union[float, Iterable, np.array]
) -> Union[float, Iterable, np.array]:
    """
    Args:
        n: years
        i: periodic interest rate in decimal form
    Returns:
        Present value of a decreasing annuity due of n years with
        an interest rate of i.  It is assumed that the annuity payment
        decrements by 1 each period.
    """
    return np.multiply(np.add(1, i), decreasing_annuity_pv(n, i))


def decreasing_annuity_due_fv(
    n: Union[int, Iterable, np.array], i: Union[float, Iterable, np.array]
) -> Union[float, Iterable, np.array]:
    """
    Args:
        n: years
        i: periodic interest rate in decimal form
    Returns:
        Future value of a decreasing annuity due of n years with
        an interest rate of i.  It is assumed that the annuity payment
        decrements by 1 each period.
    """
    return np.multiply(decreasing_annuity_due_pv(n, i), np.power(np.add(1, i), n))


def geo_increasing_annuity_pv(
    n: Union[int, Iterable, np.array],
    i: Union[int, Iterable, np.array],
    k: Union[int, Iterable, np.array],
) -> Union[float, Iterable, np.array]:
    """
    Args:
        n: years
        i: periodic interest rate in decimal form
        k: periodic payment growth rate
    Returns:
        Present value of a geometrically increasing annuity of
        n years with an interest rate of i and payment growth rate
        of k.
    """
    result = []
    n = list(n) if not isinstance(n, numbers.Number) else [n]
    i = list(i) if not isinstance(i, numbers.Number) else [i]
    k = list(k) if not isinstance(k, numbers.Number) else [k]
    for index, growth_rate in enumerate(k):
        if growth_rate == 0:
            result.append(annuity_pv(n, i))
        elif growth_rate == i[index]:
            result.append(np.multiply(n, discount_factor(i)))
        else:
            result.append(
                np.divide(
                    1 - np.power(np.divide(np.add(1, k), np.add(1, i)), n),
                    np.subtract(i, k),
                )
            )
    return np.array(result) if len(result) > 1 else result[0]



import csv
from typing import Dict, List


def read_soa_csv_mort_table(
    file_path, encoding: str = "Windows-1252", delimiter: str = ","
) -> Dict:
    """
    Args:
        file_path: The full file system path to the csv
        encoding: The text encoding of the csv data.  It defaults to 'Windows-1252'.
        delimiter: The delimiter of the csv data.  It defaults to ','.

    Returns:
        A dictionary with a 'values' and 'metadata' key.
    """
    raw_csv = _open_soa_csv_mort_table(
        file_path, encoding=encoding, delimiter=delimiter
    )
    processed_csv = _process_soa_csv_mort_table(raw_csv)
    return processed_csv


def _open_soa_csv_mort_table(
    file_path: str, encoding: str = "Windows-1252", delimiter: str = ","
) -> List[List[str]]:
    """
    Args:
        file_path: The full system path to the SOA csv table.
        url: The full HTTP url to the SOA csv table.
        encoding: The text encoding of the csv data.  It defaults to 'Windows-1252'.
        delimiter: The delimiter of the csv data.  It defaults to ','.

    Returns:
        A list of lists representing the csv table.  Each interior
        list has two elements representing the first and second
        columns.
    """
    encoded_csv = open(file_path, encoding=encoding)
    raw_csv = list(csv.reader(encoded_csv, delimiter=delimiter))
    return raw_csv


def _process_soa_csv_mort_table(raw_csv: List[List[str]]) -> Dict:
    """
    Args:
        raw_csv: A list of lists representing the csv table.  Each interior
                 list has two elements representing the first and second
                 columns.

    Returns:
        A dictionary with a 'values' and 'metadata' key.
    """
    # pylint: disable=too-many-branches
    table = {"metadata": {}, "values": ()}
    for index, row in enumerate(raw_csv):
        if row:
            if row[0] == "Row, Column (if applicable)->MinScaleValue:":
                table["metadata"]["min_age"] = int(row[1])
            elif row[0] == "Row, Column (if applicable)->MaxScaleValue:":
                table["metadata"]["max_age"] = int(row[1])
            elif row[0] == "Row\\Column":
                table["metadata"]["table_line_start"] = index + 1
            elif row[0] == "Table Name:":
                table["metadata"]["name"] = row[1]
            elif row[0] == "Table Description:":
                table["metadata"]["description"] = row[1]
            elif row[0] == "Provider Name:":
                table["metadata"]["author"] = row[1]
            elif row[0] == "Table Reference:":
                table["metadata"]["reference"] = row[1]
            elif row[0] == "Comments:":
                table["metadata"]["comments"] = row[1]
            elif row[0] == "Content Type:":
                table["metadata"]["content_type"] = row[1]
            elif row[0] == "Nation:":
                table["metadata"]["study_nation"] = row[1]
            elif row[0] == "Row, Column (if applicable)->Increment:":
                table["metadata"]["table_increment"] = row[1]
            elif row[0] == "Scaling Factor:":
                table["metadata"]["scaling_factor"] = row[1]
            elif row[0] == "Table Identity":
                table["metadata"]["soa_table_identity"] = row[1]
    table_start = table["metadata"]["table_line_start"]
    max_age = table["metadata"]["max_age"]
    min_age = table["metadata"]["min_age"]
    table_end = table_start + max_age - min_age + 1
    table["values"] = tuple([float(row[1]) for row in raw_csv[table_start:table_end]])
    return table



from typing import Callable

import makefun


class InvalidInterval(Exception):
    """
    Custom exception raised for invalid input intervals (n)
    """


class InvalidAge(Exception):
    """
    Custom exception raised for invalid input ages (x)
    """


def _get_args_dict(fn, args, kwargs):
    args_names = fn.__code__.co_varnames[: fn.__code__.co_argcount]
    return {**dict(zip(args_names, args)), **kwargs}


def validate_age(func: Callable) -> Callable:
    # pylint: disable=unused-argument
    """
    Decorator for validating methods using actuarial notation
    age input

    Args:
        func: function with inputs to validate

    Returns:
        The passed in function wrapped with input validation
    """

    @makefun.wraps(func)
    def validated_func(*args, **kwargs):
        """
        Args:
            contains an argument named 'x' for age
        """
        args_dict = _get_args_dict(func, args, kwargs)
        if args_dict["x"] < 0:
            raise InvalidAge("Start age must be greater than or equal to 0!")
        return func(*args, **kwargs)

    validated_func.__doc__ = func.__doc__
    return validated_func


def validate_interval(func: Callable) -> Callable:
    # pylint: disable=unused-argument
    """
    Decorator for validating methods using actuarial notation
    interval input

    Args:
        func: function with inputs to validate

    Returns:
        The passed in function wrapped with input validation
    """

    @makefun.wraps(func)
    def validated_func(*args, **kwargs):
        """
        Args:
            contains an argument named 'n' for interval
        """
        args_dict = _get_args_dict(func, args, kwargs)
        if args_dict["n"] <= 0:
            raise InvalidInterval("Interval must be greater than 0!")
        return func(*args, **kwargs)

    validated_func.__doc__ = func.__doc__
    return validated_func


def validate_t_interval(func: Callable) -> Callable:
    # pylint: disable=unused-argument
    """
    Decorator for validating methods using actuarial notation
    interval input

    Args:
        func: function with inputs to validate

    Returns:
        The passed in function wrapped with input validation
    """

    @makefun.wraps(func)
    def validated_func(*args, **kwargs):
        """
        Args:
            contains an argument named 't' for failure interval
        """
        args_dict = _get_args_dict(func, args, kwargs)
        if args_dict["t"] <= 0:
            raise InvalidInterval("Failure interval must be greater than 0!")
        return func(*args, **kwargs)

    validated_func.__doc__ = func.__doc__
    return validated_func



from typing import Iterable, Union

import numpy as np

from elizur.life.annuity import discount_factor


class InvalidEPVInputs(Exception):
    """
    Custom exception raised for invalid expected present value inputs.
    """


def expected_present_value(
    cash_flows: Union[Iterable, np.ndarray],
    probabilities: Union[Iterable, np.ndarray],
    interest_rates: Union[Iterable, np.ndarray],
) -> Union[float, np.ndarray]:
    """
    This function is useful for calculating variable streams of cash flows,
    interest rates, and probabilities.

    Args:
        cash_flows: payouts from time 0 to n, where time n is the last
        possible payout, e.g., (cf0, cf1, cf2, ..., cfn-1, cfn)
        probabilities: probability of a cash flow occuring from time
                       0 to n, e.g., (1p0, 2p1, 3p2, ..., npn-1)
        interest_rates: collection of interest rates to use for
                        discounting, where each interest rate is for one
                        period, e.g., (i0to1, i1to2, ..., in-1ton)

    Returns:
        The expected present value

    Example:
        Given the probabilities: 1p0 = 0.98, 2p1=0.94, 3p2=0.91

        annuity(x=0, i=0.07, n=3)

        expected_present_value((1, 1, 1), (0.98, 0.94, 0.91), (0.07, 0.07, 0.07))

        The two methods of calculating the present value of an annuity are the same

        Additionally a two-dimensional arrays of inputs can be provided to perform
        multiple expected present values at once.  The work flows below produce the
        same results:

        expected_present_value((1, 1, 1), (0.98, 0.94, 0.91), (0.07, 0.07, 0.07))

        expected_present_value((1, 1, 1), (0.88, 0.84, 0.81), (0.06, 0.06, 0.06))

        expected_present_value((1, 1, 1), (0.78, 0.74, 0.71), (0.05, 0.05, 0.05))

        expected_present_value(
            np.array([
                [1, 1, 1],
                [1, 1, 1],
                [1, 1, 1]
            ]),
            np.array([
                [0.98, 0.96, 0.91],
                [0.88, 0.86, 0.81],
                [0.78, 0.76, 0.71]
            ]),
            np.array([
                [0.07, 0.07, 0.07],
                [0.06, 0.06, 0.06],
                [0.05, 0.05, 0.05]
            ])
        )
    """
    cash_flows = np.array(cash_flows)
    probabilities = np.array(probabilities)
    interest_rates = np.array(interest_rates)

    if not cash_flows.size == probabilities.size == interest_rates.size:
        raise InvalidEPVInputs(
            "The shape of the inputs do not match! The cash "
            f"flow shape is {cash_flows.shape}, the probability "
            f"shape is {probabilities.shape}, "
            f"the interest rate shape is {interest_rates.shape}!"
        )
    if cash_flows.ndim > 2 or probabilities.ndim > 2 or interest_rates.ndim > 2:
        raise InvalidEPVInputs(
            "The dimensions of the inputs are too high! The cash "
            f"flow number of dimensions is {cash_flows.ndim}, the probability "
            f"number of dimensions is {probabilities.ndim}, "
            f"the interest rate number of dimensions is {interest_rates.ndim}! "
            "The number of dimensions for each input must be less than 3."
        )

    discount_factors = discount_factor(interest_rates)
    if interest_rates.ndim == 1:
        cumulative_product = np.nancumprod(discount_factors)
        epv = np.nansum(
            np.multiply(np.multiply(cash_flows, probabilities), cumulative_product)
        )
    else:
        cumulative_product = np.apply_along_axis(np.nancumprod, 1, discount_factors)
        epv = np.apply_along_axis(
            np.nansum,
            1,
            np.multiply(np.multiply(cash_flows, probabilities), cumulative_product),
        )
    return epv



from setuptools import setup, find_packages

from elizur import VERSION


def readme():
    with open("README.md") as f:
        return f.read()


setup(
    author="Tanner Rollefson",
    classifiers=[
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Natural Language :: English",
        "Operating System :: POSIX :: Linux",
        "Operating System :: MacOS",
        "Operating System :: Microsoft :: Windows",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Topic :: Scientific/Engineering :: Mathematics",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Financial and Insurance Industry",
    ],
    description="Elizur is an open source finance package for actuaries, finance professionals, and students",
    install_requires=["numpy>=1.18", "makefun>=1.9.2"],
    extras_require={
        "dev": [
            "black",
            "coveralls",
            "mock",
            "pylint",
            "pytest",
            "pytest-cov",
            "sphinx",
            "sphinx_rtd_theme",
        ]
    },
    keywords="actuary, actuarial, life, contingencies, finance, math, elizur",
    long_description=readme(),
    long_description_content_type="text/markdown",
    name="elizur",
    packages=find_packages(include=["elizur*"]),
    project_urls={
        "Documentation": "https://trollefson.github.io/elizur",
        "Source": "https://github.com/trollefson/elizur",
    },
    url="https://github.com/trollefson/elizur",
    version=VERSION,
    zip_safe=False,
)
